# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

import io
import warnings
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import ANY, call, Mock, patch

import black
import ruff_api
import trailrunner
import usort
from black.mode import TargetVersion
from libcst import ParserSyntaxError

import ufmt
from ufmt.core import ufmt_stdin
from ufmt.types import (
    BlackConfig,
    Encoding,
    FileContent,
    Formatter,
    Result,
    SkipFormatting,
    Sorter,
    STDIN,
    UfmtConfig,
    UsortConfig,
)

FAKE_CONFIG = """
[tool.ufmt]
excludes = [
    "foo/frob/",
    "__init__.py",
]
"""

POORLY_FORMATTED_CODE = """\
import click
from re import compile
import sys
import volatile



class Foo:
    var:  bool=123
    def method(foo,
               bar):
            '''Some docstring'''
            value = [str(some_really_long_var) for some_really_long_var in range(10) if some_really_long_var > 5]
            print(f'hello {value}!')
def func(arg:str="default")->bool:
        return arg  =="default" """  # noqa: E501

CORRECTLY_FORMATTED_CODE = '''\
import sys
from re import compile

import click
import volatile


class Foo:
    var: bool = 123

    def method(foo, bar):
        """Some docstring"""
        value = [
            str(some_really_long_var)
            for some_really_long_var in range(10)
            if some_really_long_var > 5
        ]
        print(f"hello {value}!")


def func(arg: str = "default") -> bool:
    return arg == "default"
'''

POORLY_FORMATTED_STUB = """\
import click
import os

def foo(a: int, b: str) -> bool:
    ...

class Bar:
    def hello(
        self,
    ) -> None:
        ...
"""

CORRECTLY_FORMATTED_STUB = """\
import os

import click

def foo(a: int, b: str) -> bool: ...

class Bar:
    def hello(
        self,
    ) -> None: ...
"""

INVALID_SYNTAX = """\
print "hello world"
"""


@patch.object(trailrunner.core.Trailrunner, "DEFAULT_EXECUTOR", ThreadPoolExecutor)
class CoreTest(TestCase):
    maxDiff = None

    @patch("ufmt.core.black.format_file_contents", wraps=black.format_file_contents)
    @patch("ufmt.core.ruff_api.format_string", wraps=ruff_api.format_string)
    def test_ufmt_bytes(self, ruff_mock: Mock, black_mock: Mock) -> None:
        black_config = BlackConfig()
        usort_config = UsortConfig()

        with self.subTest("changed"):
            black_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                POORLY_FORMATTED_CODE.encode(),
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), result)
            black_mock.assert_called_once()
            ruff_mock.assert_not_called()

        with self.subTest("unchanged"):
            black_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                CORRECTLY_FORMATTED_CODE.encode(),
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), result)
            black_mock.assert_called_once()
            ruff_mock.assert_not_called()

        with self.subTest("type stub"):
            black_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.pyi"),
                POORLY_FORMATTED_STUB.encode(),
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_STUB.encode(), result)
            black_mock.assert_called_once()
            ruff_mock.assert_not_called()

    @patch("ufmt.core.black.format_file_contents", wraps=black.format_file_contents)
    @patch("ufmt.core.ruff_api.format_string", wraps=ruff_api.format_string)
    def test_ufmt_bytes_alternate_formatter(
        self, ruff_mock: Mock, black_mock: Mock
    ) -> None:
        black_config = BlackConfig()
        usort_config = UsortConfig()

        with self.subTest("changed"):
            ruff_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                POORLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(formatter=Formatter.ruff_api),
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), result)
            ruff_mock.assert_called_once()
            black_mock.assert_not_called()

        with self.subTest("unchanged"):
            ruff_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                CORRECTLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(formatter=Formatter.ruff_api),
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), result)
            ruff_mock.assert_called_once()
            black_mock.assert_not_called()

        with self.subTest("type stub"):
            ruff_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.pyi"),
                POORLY_FORMATTED_STUB.encode(),
                ufmt_config=UfmtConfig(formatter=Formatter.ruff_api),
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_STUB.encode(), result)
            ruff_mock.assert_called_once()
            black_mock.assert_not_called()

        with self.subTest("target version"):
            ruff_mock.reset_mock()
            black_config_tv = BlackConfig(
                target_versions={TargetVersion.PY312},
            )
            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                POORLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(formatter=Formatter.ruff_api),
                black_config=black_config_tv,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), result)
            ruff_mock.assert_called_once()
            _, call_kwargs = ruff_mock.call_args
            self.assertEqual(call_kwargs["options"].target_version, "py312")
            # verify target_versions was not mutated
            self.assertEqual(black_config_tv.target_versions, {TargetVersion.PY312})
            black_mock.assert_not_called()

        with self.subTest("unsupported formatter"):
            with self.assertRaisesRegex(
                ValueError, "'garbage' is not a supported formatter"
            ):
                ufmt.ufmt_bytes(
                    Path("foo.pyi"),
                    POORLY_FORMATTED_STUB.encode(),
                    ufmt_config=UfmtConfig(formatter="garbage"),  # type: ignore
                    black_config=black_config,
                    usort_config=usort_config,
                )
            ruff_mock.assert_called_once()
            black_mock.assert_not_called()

    @patch("ufmt.core.usort", wraps=usort.usort)
    @patch("ufmt.core.ruff_api.isort_string", wraps=ruff_api.isort_string)
    def test_ufmt_bytes_alternate_sorter(
        self, ruff_mock: Mock, usort_mock: Mock
    ) -> None:
        black_config = BlackConfig()
        usort_config = UsortConfig()

        with self.subTest("default"):
            usort_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                POORLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(),
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), result)
            usort_mock.assert_called_once()

        with self.subTest("usort"):
            usort_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                POORLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(),
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), result)
            usort_mock.assert_called_with(
                POORLY_FORMATTED_CODE.encode(), usort_config, Path("foo.py")
            )
            ruff_mock.assert_not_called()

            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                CORRECTLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(sorter=Sorter.skip),
                black_config=black_config,
                usort_config=usort_config,
            )
            expected = CORRECTLY_FORMATTED_CODE
            self.assertEqual(expected.encode(), result)

        with self.subTest("ruff-api"):
            usort_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                POORLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(sorter=Sorter.ruff_api),
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), result)
            usort_mock.assert_not_called()
            ruff_mock.assert_called_with(
                "foo.py", POORLY_FORMATTED_CODE, options=ANY, root=None
            )

            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                CORRECTLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(sorter=Sorter.skip),
                black_config=black_config,
                usort_config=usort_config,
            )
            expected = CORRECTLY_FORMATTED_CODE
            self.assertEqual(expected.encode(), result)

        with self.subTest("skip"):
            usort_mock.reset_mock()
            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                POORLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(sorter=Sorter.skip),
                black_config=black_config,
                usort_config=usort_config,
            )
            expected = POORLY_FORMATTED_CODE[:64] + CORRECTLY_FORMATTED_CODE[65:]
            self.assertEqual(expected.encode(), result)
            usort_mock.assert_not_called()

            result = ufmt.ufmt_bytes(
                Path("foo.py"),
                CORRECTLY_FORMATTED_CODE.encode(),
                ufmt_config=UfmtConfig(sorter=Sorter.skip),
                black_config=black_config,
                usort_config=usort_config,
            )
            expected = CORRECTLY_FORMATTED_CODE
            self.assertEqual(expected.encode(), result)

        with self.subTest("unsupported sorter"):
            with self.assertRaisesRegex(
                ValueError, "'garbage' is not a supported sorter"
            ):
                ufmt.ufmt_bytes(
                    Path("foo.py"),
                    POORLY_FORMATTED_CODE.encode(),
                    ufmt_config=UfmtConfig(sorter="garbage"),  # type: ignore
                    black_config=black_config,
                    usort_config=usort_config,
                )
            usort_mock.assert_not_called()

    def test_ufmt_bytes_pre_processor(self) -> None:
        def pre_processor(
            path: Path, content: bytes, *, encoding: Encoding = "utf-8"
        ) -> bytes:
            return content + b"print('hello')\n"

        black_config = BlackConfig()
        usort_config = UsortConfig()

        result = ufmt.ufmt_bytes(
            Path("foo.py"),
            CORRECTLY_FORMATTED_CODE.encode(),
            black_config=black_config,
            usort_config=usort_config,
            pre_processor=pre_processor,
        )

        self.assertEqual(
            CORRECTLY_FORMATTED_CODE.encode() + b'\n\nprint("hello")\n', result
        )

    def test_ufmt_bytes_post_processor(self) -> None:
        def post_processor(
            path: Path, content: bytes, *, encoding: Encoding = "utf-8"
        ) -> bytes:
            return content + b"\nprint('hello')\n"

        black_config = BlackConfig()
        usort_config = UsortConfig()

        result = ufmt.ufmt_bytes(
            Path("foo.py"),
            CORRECTLY_FORMATTED_CODE.encode(),
            black_config=black_config,
            usort_config=usort_config,
            post_processor=post_processor,
        )

        self.assertEqual(
            CORRECTLY_FORMATTED_CODE.encode() + b"\nprint('hello')\n", result
        )

    def test_ufmt_string(self) -> None:
        black_config = BlackConfig()
        usort_config = UsortConfig()

        warnings.simplefilter("ignore", DeprecationWarning)

        with self.subTest("changed"):
            result = ufmt.ufmt_string(
                Path("foo.py"),
                POORLY_FORMATTED_CODE,
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE, result)

        with self.subTest("unchanged"):
            result = ufmt.ufmt_string(
                Path("foo.py"),
                CORRECTLY_FORMATTED_CODE,
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE, result)

        with self.subTest("type stub"):
            result = ufmt.ufmt_string(
                Path("foo.pyi"),
                POORLY_FORMATTED_STUB,
                black_config=black_config,
                usort_config=usort_config,
            )
            self.assertEqual(CORRECTLY_FORMATTED_STUB, result)

        warnings.resetwarnings()

        with self.subTest("version check"):
            self.assertRegex(ufmt.__version__, r"^2\.", "remove ufmt_string in 3.0")

    def test_ufmt_file(self) -> None:
        with TemporaryDirectory() as td:
            tdp = Path(td)
            f = tdp / "foo.py"
            f.write_text(POORLY_FORMATTED_CODE)

            with self.subTest("dry run"):
                result = ufmt.ufmt_file(f, dry_run=True)
                self.assertTrue(result.changed)
                self.assertIsNone(result.diff)
                self.assertEqual(POORLY_FORMATTED_CODE, f.read_text())

            with self.subTest("dry run with diff"):
                result = ufmt.ufmt_file(f, dry_run=True, diff=True)
                self.assertTrue(result.changed)
                self.assertIsNotNone(result.diff)
                self.assertEqual(POORLY_FORMATTED_CODE, f.read_text())

            with self.subTest("for reals"):
                result = ufmt.ufmt_file(f)
                self.assertTrue(result.changed)
                self.assertEqual(CORRECTLY_FORMATTED_CODE, f.read_text())

            with self.subTest("already formatted"):
                f.write_text(CORRECTLY_FORMATTED_CODE)

                result = ufmt.ufmt_file(f)
                self.assertFalse(result.changed)
                self.assertEqual(CORRECTLY_FORMATTED_CODE, f.read_text())

            with self.subTest("already formatted, unix newlines"):
                unix_content = CORRECTLY_FORMATTED_CODE.encode().replace(b"\r\n", b"\n")
                f.write_bytes(unix_content)

                result = ufmt.ufmt_file(f)
                self.assertFalse(result.changed)
                self.assertEqual(unix_content, f.read_bytes())

            with self.subTest("already formatted, windows newlines"):
                windows_content = CORRECTLY_FORMATTED_CODE.encode().replace(
                    b"\n", b"\r\n"
                )
                f.write_bytes(windows_content)

                result = ufmt.ufmt_file(f)
                self.assertFalse(result.changed)
                self.assertEqual(windows_content, f.read_bytes())

            with self.subTest("invalid syntax"):
                f.write_text(INVALID_SYNTAX)

                result = ufmt.ufmt_file(f)
                self.assertIsInstance(result.error, ParserSyntaxError)

            with self.subTest("file not found"):
                f.unlink()

                result = ufmt.ufmt_file(f)
                self.assertIsInstance(result.error, FileNotFoundError)

            with self.subTest("write failed"):
                with patch("ufmt.core.write_file") as write_mock:
                    write_mock.side_effect = PermissionError("fake permission error")

                    f.write_text(POORLY_FORMATTED_CODE)

                    result = ufmt.ufmt_file(f)
                    self.assertIsInstance(result.error, PermissionError)

            with self.subTest("skipped file no reason"):

                def skip_no_reason(
                    path: Path, content: FileContent, *, encoding: Encoding = "utf-8"
                ) -> FileContent:
                    raise SkipFormatting

                f.write_text(POORLY_FORMATTED_CODE)

                result = ufmt.ufmt_file(f, pre_processor=skip_no_reason)
                self.assertTrue(result.skipped)
                self.assertIs(result.skipped, True)
                self.assertEqual(POORLY_FORMATTED_CODE, f.read_text())

            with self.subTest("skipped file no reason"):

                def skip_with_reason(
                    path: Path, content: FileContent, *, encoding: Encoding = "utf-8"
                ) -> FileContent:
                    raise SkipFormatting("because I said so")

                f.write_text(POORLY_FORMATTED_CODE)

                result = ufmt.ufmt_file(f, pre_processor=skip_with_reason)
                self.assertTrue(result.skipped)
                self.assertEqual(result.skipped, "because I said so")
                self.assertEqual(POORLY_FORMATTED_CODE, f.read_text())

    @patch("ufmt.core.sys.stdin")
    @patch("ufmt.core.sys.stdout")
    def test_ufmt_stdin(self, stdout_mock: Mock, stdin_mock: Mock) -> None:
        with self.subTest("check"):
            stdin_mock.buffer = stdin = io.BytesIO()
            stdout_mock.buffer = stdout = io.BytesIO()

            stdin.write(POORLY_FORMATTED_CODE.encode())
            stdin.seek(0)

            result = ufmt_stdin(STDIN, dry_run=True)
            expected = Result(path=STDIN, changed=True)
            self.assertEqual(expected, result)
            stdout.seek(0)
            self.assertEqual(b"", stdout.read())

        with self.subTest("diff"):
            stdin_mock.buffer = stdin = io.BytesIO()
            stdout_mock.buffer = stdout = io.BytesIO()

            stdin.write(POORLY_FORMATTED_CODE.encode())
            stdin.seek(0)

            result = ufmt_stdin(STDIN, dry_run=True, diff=True)
            self.assertTrue(result.diff)
            stdout.seek(0)
            self.assertEqual(b"", stdout.read())

        with self.subTest("diff with path"):
            stdin_mock.buffer = stdin = io.BytesIO()
            stdout_mock.buffer = stdout = io.BytesIO()

            stdin.write(POORLY_FORMATTED_CODE.encode())
            stdin.seek(0)

            path = Path("hello/world.py")
            result = ufmt_stdin(path, dry_run=True, diff=True)
            self.assertIsNotNone(result.diff)
            self.assertRegex(
                result.diff or "", r"--- hello.world\.py\n\+\+\+ hello.world\.py"
            )
            stdout.seek(0)
            self.assertEqual(b"", stdout.read())

        with self.subTest("format"):
            stdin_mock.buffer = stdin = io.BytesIO()
            stdout_mock.buffer = stdout = io.BytesIO()

            stdin.write(POORLY_FORMATTED_CODE.encode())
            stdin.seek(0)

            result = ufmt_stdin(STDIN)
            expected = Result(path=STDIN, changed=True, written=True)
            stdout.seek(0)
            self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), stdout.read())

    def test_ufmt_paths(self) -> None:
        with TemporaryDirectory() as td:
            tdp = Path(td)
            f1 = tdp / "bar.py"
            sd = tdp / "foo"
            sd.mkdir()
            f2 = sd / "baz.py"
            f3 = sd / "frob.py"

            for f in f1, f2, f3:
                f.write_text(POORLY_FORMATTED_CODE)

            file_wrapper = Mock(name="ufmt_file", wraps=ufmt.ufmt_file)
            with patch("ufmt.core.ufmt_file", file_wrapper):
                with self.subTest("no paths"):
                    results = list(ufmt.ufmt_paths([], dry_run=True))
                    self.assertEqual([], results)
                    file_wrapper.assert_not_called()

                with self.subTest("non-existent paths"):
                    results = list(
                        ufmt.ufmt_paths(
                            [(tdp / "fake.py"), (tdp / "another.py")], dry_run=True
                        )
                    )
                    self.assertEqual([], results)

                with self.subTest("mixed paths with stdin"):
                    with patch("ufmt.core.LOG") as log_mock:
                        results = list(ufmt.ufmt_paths([f1, STDIN, f3], dry_run=True))
                        self.assertEqual(2, len(results))
                        log_mock.warning.assert_called_once()

                with self.subTest("files"):
                    results = list(ufmt.ufmt_paths([f1, f3], dry_run=True))
                    self.assertEqual(2, len(results))
                    file_wrapper.assert_has_calls(
                        [
                            call(
                                f1,
                                dry_run=True,
                                diff=False,
                                return_content=False,
                                ufmt_config_factory=None,
                                black_config_factory=None,
                                usort_config_factory=None,
                                pre_processor=None,
                                post_processor=None,
                            ),
                            call(
                                f3,
                                dry_run=True,
                                diff=False,
                                return_content=False,
                                ufmt_config_factory=None,
                                black_config_factory=None,
                                usort_config_factory=None,
                                pre_processor=None,
                                post_processor=None,
                            ),
                        ],
                        any_order=True,
                    )
                    self.assertTrue(all(r.changed for r in results))
                    file_wrapper.reset_mock()

                with self.subTest("files with diff"):
                    results = list(ufmt.ufmt_paths([f1, f3], dry_run=True, diff=True))
                    self.assertEqual(2, len(results))
                    file_wrapper.assert_has_calls(
                        [
                            call(
                                f1,
                                dry_run=True,
                                diff=True,
                                return_content=False,
                                ufmt_config_factory=None,
                                black_config_factory=None,
                                usort_config_factory=None,
                                pre_processor=None,
                                post_processor=None,
                            ),
                            call(
                                f3,
                                dry_run=True,
                                diff=True,
                                return_content=False,
                                ufmt_config_factory=None,
                                black_config_factory=None,
                                usort_config_factory=None,
                                pre_processor=None,
                                post_processor=None,
                            ),
                        ],
                        any_order=True,
                    )
                    self.assertTrue(all(r.changed for r in results))
                    file_wrapper.reset_mock()

                with self.subTest("subdir"):
                    results = list(ufmt.ufmt_paths([sd]))
                    file_wrapper.assert_has_calls(
                        [
                            call(
                                f2,
                                dry_run=False,
                                diff=False,
                                return_content=False,
                                ufmt_config_factory=None,
                                black_config_factory=None,
                                usort_config_factory=None,
                                pre_processor=None,
                                post_processor=None,
                            ),
                            call(
                                f3,
                                dry_run=False,
                                diff=False,
                                return_content=False,
                                ufmt_config_factory=None,
                                black_config_factory=None,
                                usort_config_factory=None,
                                pre_processor=None,
                                post_processor=None,
                            ),
                        ],
                        any_order=True,
                    )
                    self.assertTrue(all(r.changed for r in results))
                    file_wrapper.reset_mock()

    @patch("ufmt.core.ufmt_stdin")
    def test_ufmt_paths_stdin(self, stdin_mock: Mock) -> None:
        stdin_mock.return_value = Result(path=STDIN, changed=True)

        with self.subTest("no name"):
            list(ufmt.ufmt_paths([STDIN], dry_run=True))
            stdin_mock.assert_called_with(
                Path("stdin"),
                dry_run=True,
                diff=False,
                return_content=False,
                ufmt_config_factory=None,
                black_config_factory=None,
                usort_config_factory=None,
                pre_processor=None,
                post_processor=None,
            )

        with self.subTest("path name"):
            list(ufmt.ufmt_paths([STDIN, Path("hello.py")], dry_run=True))
            stdin_mock.assert_called_with(
                Path("hello.py"),
                dry_run=True,
                diff=False,
                return_content=False,
                ufmt_config_factory=None,
                black_config_factory=None,
                usort_config_factory=None,
                pre_processor=None,
                post_processor=None,
            )

        with self.subTest("extra args"):
            with self.assertRaisesRegex(ValueError, "too many stdin paths"):
                list(
                    ufmt.ufmt_paths(
                        [STDIN, Path("hello.py"), Path("foo.py")], dry_run=True
                    )
                )

    @patch("ufmt.core.sys.stdin")
    @patch("ufmt.core.sys.stdout")
    def test_ufmt_paths_stdin_resolves(
        self, stdout_mock: Mock, stdin_mock: Mock
    ) -> None:
        stdin_mock.buffer = io.BytesIO(POORLY_FORMATTED_CODE.encode())
        stdout_mock.buffer = stdout = io.BytesIO()

        def fake_preprocessor(
            path: Path, content: FileContent, encoding: Encoding
        ) -> FileContent:
            # ensure the fake path resolves correctly (#94)
            path.resolve()
            return content

        result = list(ufmt.ufmt_paths([STDIN]))
        expected = [Result(Path("stdin"), changed=True, written=True)]
        stdout.seek(0)
        output = stdout.read()

        self.assertListEqual(expected, result)
        self.assertEqual(CORRECTLY_FORMATTED_CODE.encode(), output)

    def test_ufmt_paths_config(self) -> None:
        with TemporaryDirectory() as td:
            tdp = Path(td).resolve()
            md = tdp / "foo"
            md.mkdir()
            f1 = md / "__init__.py"
            f2 = md / "foo.py"
            sd = md / "frob/"
            sd.mkdir()
            f3 = sd / "main.py"

            for f in f1, f2, f3:
                f.write_text(POORLY_FORMATTED_CODE)

            pyproj = tdp / "pyproject.toml"
            pyproj.write_text(FAKE_CONFIG)

            file_wrapper = Mock(name="ufmt_file", wraps=ufmt.ufmt_file)
            with patch("ufmt.core.ufmt_file", file_wrapper):
                list(ufmt.ufmt_paths([tdp]))
                file_wrapper.assert_has_calls(
                    [
                        call(
                            f2,
                            dry_run=False,
                            diff=False,
                            return_content=False,
                            ufmt_config_factory=None,
                            black_config_factory=None,
                            usort_config_factory=None,
                            pre_processor=None,
                            post_processor=None,
                        ),
                    ],
                    any_order=True,
                )
                self.assertEqual(f1.read_text(), POORLY_FORMATTED_CODE)
                self.assertEqual(f2.read_text(), CORRECTLY_FORMATTED_CODE)
                self.assertEqual(f3.read_text(), POORLY_FORMATTED_CODE)

    def test_e2e_empty_files(self) -> None:
        with TemporaryDirectory() as td:
            tdp = Path(td).resolve()
            foo = tdp / "foo.py"
            foo.write_text("")

            results = list(ufmt.ufmt_paths([foo], return_content=True))
            expected = [
                ufmt.Result(
                    foo,
                    changed=False,
                    error=None,
                    before=b"",
                    after=b"",
                )
            ]
            self.assertListEqual(expected, results)

    def test_e2e_return_bytes(self) -> None:
        with TemporaryDirectory() as td:
            tdp = Path(td).resolve()
            foo = tdp / "foo.py"

            with self.subTest("unix newlines"):
                foo.write_bytes(POORLY_FORMATTED_CODE.encode())

                results = list(ufmt.ufmt_paths([foo], return_content=True))
                expected = [
                    ufmt.Result(
                        foo,
                        changed=True,
                        written=True,
                        diff=None,
                        error=None,
                        before=POORLY_FORMATTED_CODE.encode(),
                        after=CORRECTLY_FORMATTED_CODE.encode(),
                    )
                ]
                self.assertEqual(expected, results)

            with self.subTest("crlf newlines"):
                before = POORLY_FORMATTED_CODE.replace("\n", "\r\n").encode()
                after = CORRECTLY_FORMATTED_CODE.replace("\n", "\r\n").encode()

                foo.write_bytes(before)

                results = list(ufmt.ufmt_paths([foo], return_content=True))
                expected = [
                    ufmt.Result(
                        foo,
                        changed=True,
                        written=True,
                        diff=None,
                        error=None,
                        before=before,
                        after=after,
                    )
                ]
                self.assertEqual(expected, results)



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

import logging
import sys
from pathlib import Path
from typing import Iterable, List, Optional, Tuple

import click
from moreorless.click import echo_color_precomputed_diff

from .__version__ import __version__
from .core import ufmt_paths
from .types import Options, Result
from .util import enable_libcst_native


def init_logging(*, debug: Optional[bool] = None) -> None:
    format = "%(message)s" if not debug else "%(levelname)s %(name)s %(message)s"
    level = (
        logging.DEBUG if debug else (logging.INFO if debug is None else logging.ERROR)
    )
    logging.basicConfig(stream=sys.stderr, level=level, format=format)
    logging.getLogger("blib2to3").setLevel(logging.WARNING)


def echo_results(
    results: Iterable[Result], diff: bool = False, quiet: bool = False
) -> Tuple[int, int]:
    empty = True
    error = 0
    changed = 0
    written = 0
    clean = 0

    for result in results:
        empty = False

        if result.error is not None:
            msg = str(result.error)
            lines = msg.splitlines()
            msg = lines[0] if lines else repr(result.error)
            click.secho(f"Error formatting {result.path}: {msg}", fg="yellow", err=True)
            error += 1

        elif result.skipped:
            reason = f": {result.skipped}" if isinstance(result.skipped, str) else ""
            if not quiet:
                click.secho(f"Skipped {result.path}{reason}", err=True)

        elif result.changed:
            if result.written:
                written += 1
                if not quiet:
                    click.secho(f"Formatted {result.path}", err=True)
            else:
                changed += 1
                click.secho(f"Would format {result.path}", err=True)
            if diff and result.diff:
                echo_color_precomputed_diff(result.diff)

        else:
            clean += 1

    if not quiet:

        def f(v: int, word: str = "file") -> str:
            return f"{v} {word if v == 1 else word + 's'}"

        reports = []
        if error:
            reports += [click.style(f(error, "error"), fg="yellow", bold=True)]
        if changed:
            reports += [click.style(f"{f(changed)} would be formatted", bold=True)]
        if written:
            reports += [click.style(f"{f(written)} formatted")]
        if clean:
            reports += [click.style(f"{f(clean)} already formatted")]

        if empty:
            click.secho("❗️ No files found ❗️", fg="yellow", err=True)
        else:
            message = ", ".join(reports)
            click.secho(f"✨ {message} ✨", err=True)

    return (changed + written), error


@click.group()
@click.pass_context
@click.version_option(__version__, "--version", "-V")
@click.option(
    "--debug/--quiet",
    "-v/-q",
    is_flag=True,
    default=None,
    help="Enable debug/verbose output",
)
@click.option(
    "--concurrency",
    type=int,
    default=None,
    help="Override the default concurrency",
)
@click.option(
    "--root",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    default=None,
    help="Specify the root directory for project configuration",
)
def main(
    ctx: click.Context,
    debug: Optional[bool],
    concurrency: Optional[int],
    root: Optional[Path],
) -> None:
    init_logging(debug=debug)

    ctx.obj = Options(
        debug=debug is True,
        quiet=debug is False,
        concurrency=concurrency,
        root=root,
    )
    enable_libcst_native()


@main.command()
@click.pass_context
@click.argument(
    "names", type=click.Path(allow_dash=True), nargs=-1, metavar="[PATH] ..."
)
def check(ctx: click.Context, names: List[str]) -> None:
    """Check formatting of one or more paths"""
    options: Options = ctx.obj
    paths = [Path(name) for name in names] if names else [Path(".")]
    results = ufmt_paths(
        paths, dry_run=True, concurrency=options.concurrency, root=options.root
    )
    changed, error = echo_results(results, quiet=options.quiet)
    if changed or error:
        ctx.exit(1)


@main.command()
@click.pass_context
@click.argument(
    "names", type=click.Path(allow_dash=True), nargs=-1, metavar="[PATH] ..."
)
def diff(ctx: click.Context, names: List[str]) -> None:
    """Generate diffs for any files that need formatting"""
    options: Options = ctx.obj
    paths = [Path(name) for name in names] if names else [Path(".")]
    results = ufmt_paths(
        paths,
        dry_run=True,
        diff=True,
        concurrency=options.concurrency,
        root=options.root,
    )
    changed, error = echo_results(results, diff=True, quiet=options.quiet)
    if changed or error:
        ctx.exit(1)


@main.command()
@click.pass_context
@click.argument(
    "names", type=click.Path(allow_dash=True), nargs=-1, metavar="[PATH] ..."
)
def format(ctx: click.Context, names: List[str]) -> None:
    """Format one or more paths in place"""
    options: Options = ctx.obj
    paths = [Path(name) for name in names] if names else [Path(".")]
    results = ufmt_paths(paths, concurrency=options.concurrency, root=options.root)
    _, error = echo_results(results, quiet=options.quiet)
    if error:
        ctx.exit(1)


@main.command()
@click.pass_context
@click.option("--tcp", is_flag=True)
@click.option("--ws", is_flag=True)
@click.option("--port", type=int, default=8971)
def lsp(ctx: click.Context, tcp: bool, ws: bool, port: int) -> None:
    """Experimental: start an LSP formatting server"""
    from .lsp import ufmt_lsp

    options: Options = ctx.obj
    server = ufmt_lsp(root=options.root)

    if tcp:
        server.start_tcp("localhost", port)
    elif ws:
        server.start_ws("localhost", port)
    else:
        server.start_io()



# Copyright Amethyst Reese, Tim Hatch
# Licensed under the MIT license

import time
from pathlib import Path
from typing import Any, List

from typing_extensions import Self

from ufmt import ufmt_file, ufmt_paths

ROOT = Path(__file__).parent.parent.parent


class Benchmark:
    def __init__(self, target: int = 5) -> None:
        self.target = target
        self.last_time: int = 0
        self.title: str = ""
        self.totals: List[int] = []

    @classmethod
    def fields(self) -> str:
        headline = f"{'benchmark':^40}  {'min':^10}  {'mean':^10}  {'max':^10}"
        underline = "-" * len(headline)
        return f"{headline}\n{underline}"

    def results(self) -> str:
        totals = self.totals
        short = min(totals)
        long = max(totals)
        avg = sum(totals) // len(totals)
        fields = "  ".join(f"{value // 1000:>7} µs" for value in (short, avg, long))
        return f"{self.title + ':':<40}  {fields}"

    def __enter__(self) -> Self:
        print()
        print(self.fields())
        return self

    def __exit__(self, *args: Any) -> None:
        print()

    def __call__(self, name: str) -> Self:
        self.title = name
        return self

    def __iter__(self) -> Self:
        self.last_time = 0
        self.totals = []
        return self

    def __next__(self) -> None:
        now = time.monotonic_ns()
        if self.last_time > 0:
            self.totals += [now - self.last_time]
        if len(self.totals) >= self.target:
            print(self.results())
            self.totals = []
            self.title = ""

            raise StopIteration
        self.last_time = time.monotonic_ns()


def benchmark() -> None:
    ufmt_dir = ROOT / "ufmt"
    ufmt_core = ufmt_dir / "core.py"
    assert ufmt_dir.is_dir(), f"{ufmt_dir} not found, must run benchmark from repo"

    with Benchmark() as benchmark:

        for _ in benchmark("ufmt_file"):
            ufmt_file(ufmt_core, dry_run=True)

        for _ in benchmark("ufmt_paths"):
            list(ufmt_paths([ufmt_dir], dry_run=True))


if __name__ == "__main__":
    benchmark()



# Copyright 2022 Amethyst Reese, Tim Hatch
# Licensed under the MIT license

import logging
from pathlib import Path
from typing import List, Literal, Optional

from lsprotocol.types import (
    DocumentFormattingParams,
    MessageType,
    Position,
    Range,
    ShowMessageParams,
    TEXT_DOCUMENT_FORMATTING,
    TextEdit,
)
from pygls.lsp.server import LanguageServer
from pygls.workspace import TextDocument

from .__version__ import __version__
from .config import load_config
from .core import ufmt_bytes
from .types import (
    BlackConfigFactory,
    Encoding,
    FileContent,
    Processor,
    Result,
    SkipFormatting,
    UfmtConfigFactory,
    UsortConfig,
    UsortConfigFactory,
)
from .util import make_black_config, normalize_content, normalize_result

ServerType = Literal["stdin", "tcp", "ws"]

LOG = logging.getLogger(__name__)


def _wrap_ufmt_bytes(  # pragma: nocover
    path: Path,
    content: FileContent,
    *,
    encoding: Encoding,
    ufmt_config_factory: Optional[UfmtConfigFactory] = None,
    black_config_factory: Optional[BlackConfigFactory] = None,
    usort_config_factory: Optional[UsortConfigFactory] = None,
    pre_processor: Optional[Processor] = None,
    post_processor: Optional[Processor] = None,
    root: Optional[Path] = None,
) -> Result:
    try:
        ufmt_config = (ufmt_config_factory or load_config)(path, root)
        black_config = (black_config_factory or make_black_config)(path)
        usort_config = (usort_config_factory or UsortConfig.find)(path)

        result = Result(path)

        dst_contents = ufmt_bytes(
            path,
            content,
            encoding=encoding,
            ufmt_config=ufmt_config,
            black_config=black_config,
            usort_config=usort_config,
            pre_processor=pre_processor,
            post_processor=post_processor,
        )
        result.after = dst_contents

    except SkipFormatting as e:
        result.after = content
        result.skipped = str(e) or True
        return result

    except Exception as e:
        result.error = e
        return result

    return result


def ufmt_lsp(  # pragma: nocover
    *,
    ufmt_config_factory: Optional[UfmtConfigFactory] = None,
    black_config_factory: Optional[BlackConfigFactory] = None,
    usort_config_factory: Optional[UsortConfigFactory] = None,
    pre_processor: Optional[Processor] = None,
    post_processor: Optional[Processor] = None,
    root: Optional[Path] = None,
) -> LanguageServer:
    """
    Prepare an LSP server instance.

    Keyword arguments have the same semantics as :func:`ufmt_paths`.

    Returns a LanguageServer object. User must call the ``start_io()`` or
    ``start_tcp()`` method with appropriate arguments to actually start the LSP server.
    """
    server = LanguageServer("ufmt-lsp", __version__)

    @server.feature(TEXT_DOCUMENT_FORMATTING)
    def lsp_format_document(
        ls: LanguageServer, params: DocumentFormattingParams
    ) -> Optional[List[TextEdit]]:
        document: TextDocument = ls.workspace.get_text_document(
            params.text_document.uri
        )
        path = Path(document.path).resolve()

        # XXX: we're assuming everything is UTF-8 because LSP doesn't track encodings...
        encoding: Encoding = "utf-8"
        content, newline = normalize_content(document.source.encode(encoding))

        result = _wrap_ufmt_bytes(
            path,
            content,
            encoding=encoding,
            ufmt_config_factory=ufmt_config_factory,
            black_config_factory=black_config_factory,
            usort_config_factory=usort_config_factory,
            pre_processor=pre_processor,
            post_processor=post_processor,
            root=root,
        )

        if result.error:
            ls.window_show_message(
                ShowMessageParams(
                    MessageType.Error, f"Formatting failed: {str(result.error)}"
                )
            )
            return []

        return [
            TextEdit(
                Range(
                    Position(0, 0),
                    Position(len(document.lines), 0),
                ),
                normalize_result(result.after, newline).decode(encoding),
            ),
        ]

    return server



# Copyright 2022 Amethyst Reese, Tim Hatch
# Licensed under the MIT license

import os
import tokenize
from pathlib import Path
from typing import Optional, Tuple

from black.files import find_pyproject_toml, parse_pyproject_toml
from black.mode import TargetVersion

from .types import BlackConfig, Encoding, FileContent, Newline


def make_black_config(path: Path) -> BlackConfig:
    """
    Generate a basic black config object for the given path.
    """
    config_file = find_pyproject_toml((str(path),))
    if not config_file:
        return BlackConfig()

    config = parse_pyproject_toml(config_file)

    # manually patch options that do not have a 1-to-1 match in Mode arguments
    config["target_versions"] = {
        TargetVersion[val.upper()] for val in config.pop("target_version", [])
    }
    config["string_normalization"] = not config.pop("skip_string_normalization", False)

    names = {field.name for field in BlackConfig.__dataclass_fields__.values()}
    config = {name: value for name, value in config.items() if name in names}

    return BlackConfig(**config)


def normalize_content(
    content: FileContent, newline: Optional[Newline] = None
) -> Tuple[FileContent, Newline]:
    """
    Detect bytes content line ending style and convert to UNIX style.

    No-op if ``newline`` is detected or given as ``b"\\n"``.
    """
    if newline is None:
        newline = b"\r\n" if content.find(b"\r\n", 0, 1000) > -1 else b"\n"
    if newline != b"\n":
        content = content.replace(newline, b"\n")
    return content, newline


def normalize_result(content: FileContent, newline: Newline) -> FileContent:
    """
    Convert bytes content with UNIX style line endings to the given style.

    No-op if ``newline`` is given as ``b"\\n"``.
    """
    if newline != b"\n":
        content = content.replace(b"\n", newline)
    return content


def read_file(path: Path) -> Tuple[FileContent, Encoding, Newline]:
    """
    Read a file from disk, detect encoding, and normalize newlines.

    Returns a tuple of file contents, file encoding, and original newline style.
    File contents are in bytes, with newlines normalized to UNIX style `"\\n"`.
    File encoding is detected either by PEP 263 coding line or encoding cookie.
    Newline is detected from the first line ending, and assumes that all lines in the
    file will have (or should have) consistent line endings.

    Intended for use with :func:`write_file`, where newlines get converted back to the
    original newline style before being written to disk.
    """
    with open(path, "rb") as buf:
        encoding, lines = tokenize.detect_encoding(buf.readline)
        newline = b"\r\n" if lines and lines[0].endswith(b"\r\n") else b"\n"

        buf.seek(0)
        content = buf.read()
        content, newline = normalize_content(content, newline)
        return content, encoding, newline


def write_file(path: Path, content: FileContent, newline: Newline) -> None:
    """
    Write to a file on disk, normalizing to the given newline style.

    Expects given content in bytes, with UNIX style line endings. UNIX style newlines
    will be converted to the given newline style before writing bytes to disk.

    Intended for use with content and newline style from :func:`read_file`, so that
    newlines are correctly normalized back to their original style when writing to disk.
    """
    content = normalize_result(content, newline)
    with open(path, "wb") as buf:
        buf.write(content)


def enable_libcst_native() -> None:
    """
    Enable libcst native parser if available

    TODO: replace with usort.util.enable_libcst_native or similar
    """
    try:
        import libcst.native  # noqa: F401

        os.environ["LIBCST_PARSER_TYPE"] = "native"
    except ImportError:  # pragma: nocover
        pass



# Copyright 2022 Amethyst Reese, Tim Hatch
# Licensed under the MIT license

import logging
import re
import sys
from dataclasses import replace
from functools import partial
from itertools import chain
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator, Match, Optional, Sequence
from warnings import warn

import black
import black.mode
import black.report
from moreorless import unified_diff
from trailrunner import Trailrunner
from usort import usort
from usort.config import CAT_FIRST_PARTY, CAT_STANDARD_LIBRARY

try:
    import ruff_api
except ImportError:  # pragma: nocover
    pass

from .config import load_config
from .types import (
    BlackConfig,
    BlackConfigFactory,
    Encoding,
    FileContent,
    Formatter,
    Processor,
    Result,
    SkipFormatting,
    Sorter,
    STDIN,
    UfmtConfig,
    UfmtConfigFactory,
    UsortConfig,
    UsortConfigFactory,
)
from .util import make_black_config, normalize_result, read_file, write_file

LOG = logging.getLogger(__name__)


def ufmt_bytes(
    path: Path,
    content: FileContent,
    *,
    encoding: Encoding = "utf-8",
    black_config: BlackConfig,
    usort_config: UsortConfig,
    ufmt_config: UfmtConfig = UfmtConfig(),
    pre_processor: Optional[Processor] = None,
    post_processor: Optional[Processor] = None,
) -> FileContent:
    """
    Format arbitrary bytes for the given path.

    Requires passing valid config objects for both black and µsort. If the given path
    represents a type stub (has a ``.pyi`` suffix), the black config object will be
    updated to set ``is_pyi = True``.

    This function will not catch any errors during formatting, except for "everything is
    fine" messages, like :exc:`black.NothingChanged`. All other errors must be handled
    by the code calling this function; see :func:`ufmt_file` for example error handling.

    Optionally takes a pre- or post-processor matching the :class:`Processor` protocol.
    If given, the pre-processor will be called with the original byte string content
    before it has been run through µsort or black. The return value of the pre-processor
    will be used in place of the original content when formatting.
    If given, the post processor will be called with the updated byte string content
    after it has been run through µsort and black. The return value of the post
    processor will replace the final return value of :func:`ufmt_bytes`.

    **Experimental:**
    The optional ``ufmt_config`` parameter allows passing alternative
    :class:`~ufmt.types.Formatter` and :class:`~ufmt.types.Sorter`
    values, to choose formatting and sorting implementations instead of black and µsort.
    This should be passed by alternate frontends, and be based on the project's
    configuration in ``pyproject.toml``. Optionally can be passed through from
    :func:`ufmt_file` or :func:`ufmt_paths` as ``ufmt_config_factory``.

    .. note::
        Content will be decoded before passing to black, and re-encoded to bytes
        again. Be sure to pass a known-valid unicode encoding for the content.

        Specifying the correct encoding is important! µfmt cannot do the right thing
        without the correct encoding if files contain PEP 263 coding lines, or byte
        values unrepresentable in UTF-8 (like ``\\ud800``).

        **When in doubt, use :func:`ufmt_file` or :func:`ufmt_paths`.**

        :func:`ufmt.util.read_file` can be used to both read bytes from disk, and make a
        best guess at file encodings. Otherwise, use :func:`tokenize.detect_encodings`.

    """
    if pre_processor is not None:
        content = pre_processor(path, content, encoding=encoding)

    # run configured sorter
    if ufmt_config.sorter == Sorter.usort:
        result = usort(content, usort_config, path)
        if result.error:
            raise result.error
        content = result.output
    elif ufmt_config.sorter == Sorter.ruff_api:
        ruff_isort_options = ruff_api.SortOptions(
            # match µsort configuration
            first_party_modules=[
                imp
                for imp, category in usort_config.known.items()
                if category == CAT_FIRST_PARTY
            ],
            standard_library_modules=[
                imp
                for imp, category in usort_config.known.items()
                if category == CAT_STANDARD_LIBRARY
            ],
            detect_same_package=usort_config.first_party_detection,
            # match µsort behavior for now
            case_sensitive=False,
            combine_as_imports=True,
            order_by_type=False,
        )
        content_str = ruff_api.isort_string(
            path.as_posix(),
            content.decode(encoding),
            options=ruff_isort_options,
            root=(
                ufmt_config.project_root.as_posix()
                if ufmt_config.project_root
                else None
            ),
        )
        content = content_str.encode(encoding)
    elif ufmt_config.sorter == Sorter.skip:
        pass
    else:
        raise ValueError(f"{ufmt_config.sorter!r} is not a supported sorter")

    if path.suffix == ".pyi":
        black_config = replace(black_config, is_pyi=True)

    # run configured formatter
    content_str = content.decode(encoding)
    if ufmt_config.formatter == Formatter.black:
        try:
            content_str = black.format_file_contents(
                content_str,
                fast=False,
                mode=black_config,
            )
        except black.report.NothingChanged:
            pass
    elif ufmt_config.formatter == Formatter.ruff_api:
        target_version = None
        if black_config.target_versions:
            tv = next(iter(black_config.target_versions))
            target_version = tv.name.lower()  # "PY312" -> "py312"
        ruff_format_options = ruff_api.FormatOptions(
            target_version=target_version,
            line_width=black_config.line_length,
            preview=black_config.preview,
        )
        content_str = ruff_api.format_string(
            path.as_posix(), content_str, options=ruff_format_options
        )
    else:
        raise ValueError(f"{ufmt_config.formatter!r} is not a supported formatter")
    content = content_str.encode(encoding)

    if post_processor is not None:
        content = post_processor(path, content, encoding=encoding)

    return content


def ufmt_string(
    path: Path,
    content: str,
    usort_config: UsortConfig,
    black_config: Optional[BlackConfig] = None,
    encoding: Encoding = "utf-8",
) -> str:
    """
    Format an arbitrary string value for the given path.

    .. warning::
        Specifying the correct encoding is important! µfmt cannot do the right thing
        without the correct encoding if files contain byte values unrepresentable
        in UTF-8 (like ``\\ud800``) or PEP 263 coding lines.

        When in doubt, use :func:`ufmt_file`.

    .. deprecated:: 2.0
        Will be removed in µfmt version 3.0. Use :func:`ufmt_bytes` instead.
    """
    warn(
        "ufmt_string will be removed in version 3.0; use ufmt_bytes instead",
        DeprecationWarning,
        stacklevel=2,
    )

    black_config = black_config or BlackConfig()

    data = content.encode(encoding)
    data = ufmt_bytes(
        path,
        data,
        encoding=encoding,
        ufmt_config=load_config(path),
        black_config=black_config,
        usort_config=usort_config,
    )
    return data.decode(encoding)


def ufmt_file(
    path: Path,
    *,
    dry_run: bool = False,
    diff: bool = False,
    return_content: bool = False,
    ufmt_config_factory: Optional[UfmtConfigFactory] = None,
    black_config_factory: Optional[BlackConfigFactory] = None,
    usort_config_factory: Optional[UsortConfigFactory] = None,
    pre_processor: Optional[Processor] = None,
    post_processor: Optional[Processor] = None,
    root: Optional[Path] = None,
) -> Result:
    """
    Format a single file on disk, and returns a :class:`Result`.

    Passing ``dry_run = True`` will only format the file in memory, without writing
    changes to disk. Passing ``diff = True`` will generate a unified diff of changes
    on :attr:`Result.diff`. Passing ``return_content = True`` will also populate
    :attr:`Result.before` and :attr:`Result.after` with the bytes content of the file
    from before and after formatting, respectively.

    Any errors that occur during formatting will be caught, and those exceptions will
    be attached to the :attr:`Result.error` property of the result object. It is the
    responsibility of code calling this function to check for errors in results and
    handle or surface them appropriately.

    Optionally takes ``black_config_factory`` or ``usort_config_factory`` to override
    the default configuration detection for each respective tool. Factory functions
    must take a :class:`pathlib.Path` object and return a valid :class:`BlackConfig`
    or :class:`UsortConfig` object respectively.

    Optionally takes a pre- or post-processor matching the :class:`Processor` protocol.
    If given, the pre-processor will be called with the original byte string content
    before it has been run through µsort or black. The return value of the pre-processor
    will be used in place of the original content when formatting.
    If given, the post processor will be called with the updated byte string content
    after it has been run through µsort and black. The return value of the post
    processor will replace the final return value of :func:`ufmt_bytes`.

    Raising :exc:`ufmt.SkipFormatting` from a pre- or post-processor will result in
    "skipping" the file currently being formatted. The final contents will be unchanged,
    and the :attr:`Result.skipped` attribute will be set with the string message from
    the skip exception, or ``True`` if no message is given.
    """
    path = path.resolve()
    ufmt_config = (ufmt_config_factory or load_config)(path, root)
    black_config = (black_config_factory or make_black_config)(path)
    usort_config = (usort_config_factory or UsortConfig.find)(path)

    LOG.debug(f"Checking {path}")

    result = Result(path)

    try:
        src_contents, encoding, newline = read_file(path)
        dst_contents = ufmt_bytes(
            path,
            src_contents,
            encoding=encoding,
            ufmt_config=ufmt_config,
            black_config=black_config,
            usort_config=usort_config,
            pre_processor=pre_processor,
            post_processor=post_processor,
        )
    except SkipFormatting as e:
        dst_contents = src_contents
        result.skipped = str(e) or True
        return result

    except Exception as e:
        result.error = e
        return result

    if return_content or diff:
        src_result = normalize_result(src_contents, newline)
        dst_result = normalize_result(dst_contents, newline)

        if return_content:
            result.before = src_result
            result.after = dst_result

    if src_contents != dst_contents:
        result.changed = True

        if diff:
            result.diff = unified_diff(
                src_result.decode(encoding),
                dst_result.decode(encoding),
                path.as_posix(),
            )

        if not dry_run:
            LOG.debug(f"Formatted {path}")
            try:
                write_file(path, dst_contents, newline=newline)
                result.written = True
            except Exception as e:
                result.error = e

    return result


def ufmt_stdin(
    path: Path,
    *,
    dry_run: bool = False,
    diff: bool = False,
    return_content: bool = False,
    ufmt_config_factory: Optional[UfmtConfigFactory] = None,
    black_config_factory: Optional[BlackConfigFactory] = None,
    usort_config_factory: Optional[UsortConfigFactory] = None,
    pre_processor: Optional[Processor] = None,
    post_processor: Optional[Processor] = None,
) -> Result:
    """
    Wrapper around :func:`ufmt_file` for formatting content from STDIN.

    If given ``dry_run = False``, the resulting formatted content will be printed
    to STDOUT. Diff content and changed status will be returned as part of the
    :class:`Result` object as normal.

    Requires passing a path that represents the filesystem location matching the
    contents to be formatted. Reads bytes from STDIN until EOF, writes the content to
    a temporary location on disk, and formats that file on disk using :func:`ufmt_file`.
    The :class:`Result` object will be updated to match the location given by ``path``.

    See :func:`ufmt_file` for details on parameters, config factories,
    and post processors. All parameters are passed through to :func:`ufmt_file`.
    """
    ufmt_config = (ufmt_config_factory or load_config)(path, None)
    black_config = (black_config_factory or make_black_config)(path)
    usort_config = (usort_config_factory or UsortConfig.find)(path)

    with TemporaryDirectory() as td:
        tdp = Path(td).resolve()
        temp_path = tdp / path.name

        # read from stdin
        content = sys.stdin.buffer.read()
        temp_path.write_bytes(content)

        result = ufmt_file(
            temp_path,
            dry_run=dry_run,
            diff=diff,
            return_content=return_content,
            ufmt_config_factory=lambda _path, _root: ufmt_config,
            black_config_factory=lambda _path: black_config,
            usort_config_factory=lambda _path: usort_config,
            pre_processor=pre_processor,
            post_processor=post_processor,
        )
        result.path = path
        if result.diff and path != STDIN:
            real_path_str = path.as_posix()

            def replacement(match: Match[str]) -> str:
                return match.group(1) + real_path_str

            pattern = re.compile(r"^((?:---|\+\+\+)\s+).+$", re.M)
            result.diff = pattern.sub(replacement, result.diff)

        # write to stdout if not check/diff mode
        if not dry_run:
            content = temp_path.read_bytes()
            sys.stdout.buffer.write(content)
            sys.stdout.buffer.flush()

        return result


def ufmt_paths(
    paths: Sequence[Path],
    *,
    dry_run: bool = False,
    diff: bool = False,
    return_content: bool = False,
    ufmt_config_factory: Optional[UfmtConfigFactory] = None,
    black_config_factory: Optional[BlackConfigFactory] = None,
    usort_config_factory: Optional[UsortConfigFactory] = None,
    pre_processor: Optional[Processor] = None,
    post_processor: Optional[Processor] = None,
    concurrency: Optional[int] = None,
    root: Optional[Path] = None,
) -> Generator[Result, None, None]:
    """
    Format one or more paths, recursively, ignoring any files excluded by configuration.

    Uses trailrunner to first walk all paths, and then to run :func:`ufmt_file` on each
    matching file found. If more than one eligible file is discovered after walking the
    given paths, all files will be formatted using a process pool for improved
    performance and CPU utilization.

    Returns a generator yielding :class:`Result` objects for each file formatted.
    Any errors that occur during formatting will be caught, and those exceptions will
    be attached to the :attr:`Result.error` property of the result object. It is the
    responsibility of code calling this function to check for errors in results and
    handle or surface them appropriately.

    If the first given path is STDIN (``Path("-")``), then content will be formatted
    from STDIN using :func:`ufmt_stdin`. Results will be printed to STDOUT.
    A second path argument may be given, which represents the original content's true
    path name, and will be used when printing status messages, diffs, or errors.
    Any further path names will result in a runtime error.

    See :func:`ufmt_file` for details on parameters, config factories,
    and post processors. All parameters are passed through to :func:`ufmt_file`.

    .. note::
        Factory and post processing functions must be pickleable when using
        :func:`ufmt_paths`.
    """
    if not paths:
        return

    # format stdin and short-circuit
    if paths[0] == STDIN:
        if len(paths) > 2:
            raise ValueError("too many stdin paths")
        elif len(paths) == 2:
            _, path = paths
        else:
            path = Path("stdin")
        yield ufmt_stdin(
            path,
            dry_run=dry_run,
            diff=diff,
            return_content=return_content,
            ufmt_config_factory=ufmt_config_factory,
            black_config_factory=black_config_factory,
            usort_config_factory=usort_config_factory,
            pre_processor=pre_processor,
            post_processor=post_processor,
        )
        return

    runner = (
        Trailrunner() if concurrency is None else Trailrunner(concurrency=concurrency)
    )

    def generate_paths() -> Generator[Path, None, None]:
        """
        yield paths to format, using trailrunner to walk directories and exclude paths
        """
        for path in paths:
            if path == STDIN:
                LOG.warning("Cannot mix stdin ('-') with normal paths, ignoring")
                continue
            config = (ufmt_config_factory or load_config)(path, root)
            yield from runner.walk(path, excludes=config.excludes)

    fn = partial(
        ufmt_file,
        dry_run=dry_run,
        diff=diff,
        return_content=return_content,
        ufmt_config_factory=ufmt_config_factory,
        black_config_factory=black_config_factory,
        usort_config_factory=usort_config_factory,
        pre_processor=pre_processor,
        post_processor=post_processor,
    )

    # we have a generator that yields paths to format, but for large lists of paths
    # or large directories, we want to start formatting files as soon as possible rather
    # than waiting to finish generating the entire set of paths. we also want to skip
    # using a process pool for the base case where we only need to format a single file.
    #
    # we do this here by manually pumping the generator twice, and either exit early
    # if the first pump yields nothing (no paths to format), or immediately format
    # the first item in the same process if the second pump yields nothing (only one
    # path to format). Otherwise, we fall through and pass the first and second items,
    # plus the entire rest of the generator, to trailrunner to format everything using
    # the process pool, yielding more paths from the generator on the main process
    # while the child processes start formatting paths.

    gen = generate_paths()

    try:
        first = next(gen)
    except StopIteration:  # no paths to format
        return

    try:
        second = next(gen)
    except StopIteration:  # only one path to format
        # skip multiprocessing for a single path
        yield fn(first)
        return

    # multiple paths to format
    combined = chain([first, second], gen)  # combine first, second, and the rest
    for _, result in runner.run_iter(combined, fn):
        yield result



# Copyright Amethyst Reese
# Licensed under the MIT license

from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory
from threading import Timer
from unittest import TestCase

from ufmt.lsp import ufmt_lsp


class LspTest(TestCase):
    def setUp(self) -> None:
        self.td = TemporaryDirectory()
        self.tdp = Path(self.td.name).resolve()
        self.addCleanup(self.td.cleanup)

    def test_startup_shutdown(self) -> None:
        stdin = BytesIO()
        stdout = BytesIO()

        server = ufmt_lsp(root=self.tdp)

        def times_up() -> None:
            server.shutdown()  # type: ignore[no-untyped-call]
            raise RuntimeError("had to forcefully shutdown lsp")

        timer = Timer(5, times_up)
        timer.start()
        server.start_io(stdin, stdout)
        timer.cancel()



# Copyright 2022 Amethyst Reese, Tim Hatch
# Licensed under the MIT license

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import cast
from unittest import TestCase

import tomlkit
from black.mode import TargetVersion

import ufmt
from .core import FAKE_CONFIG, POORLY_FORMATTED_CODE

FAKE_CONTENT = b"""\
import foo

def bar():
    print("hello world")
"""


class UtilTest(TestCase):
    def test_black_config(self) -> None:
        doc = tomlkit.parse(FAKE_CONFIG)
        black = tomlkit.table()
        black["target_version"] = ["py36", "py37"]
        black["skip_string_normalization"] = True
        black["line_length"] = 87
        cast(tomlkit.container.Container, doc["tool"]).add("black", black)

        with TemporaryDirectory() as td:
            tdp = Path(td)

            pyproj = tdp / "pyproject.toml"
            pyproj.write_text(tomlkit.dumps(doc))

            f = tdp / "foo.py"
            f.write_text(POORLY_FORMATTED_CODE)

            result = ufmt.ufmt_file(f, dry_run=True)
            self.assertTrue(result.changed)

            mode = ufmt.util.make_black_config(tdp)

            with self.subTest("target_versions"):
                self.assertEqual(
                    mode.target_versions,
                    {TargetVersion[item.upper()] for item in ("py36", "py37")},
                )

            with self.subTest("string_normalization"):
                self.assertFalse(mode.string_normalization)

            with self.subTest("line_length"):
                self.assertEqual(mode.line_length, 87)

    def test_normalize_content(self) -> None:
        cases = (
            (b"hello world", None, (b"hello world", b"\n")),
            (b"hello\nworld\n", None, (b"hello\nworld\n", b"\n")),
            (b"hello\r\nworld\r\n", None, (b"hello\nworld\n", b"\r\n")),
            (b"hello world", b"\n", (b"hello world", b"\n")),
            (b"hello\nworld\n", b"\n", (b"hello\nworld\n", b"\n")),
            (b"hello\r\nworld\r\n", b"\n", (b"hello\r\nworld\r\n", b"\n")),
            (b"hello world", b"\r\n", (b"hello world", b"\r\n")),
            (b"hello\nworld\n", b"\r\n", (b"hello\nworld\n", b"\r\n")),
            (b"hello\r\nworld\r\n", b"\r\n", (b"hello\nworld\n", b"\r\n")),
        )
        for idx, (content, newline, expected) in enumerate(cases):
            with self.subTest(idx):
                self.assertEqual(
                    expected, ufmt.util.normalize_content(content, newline)
                )

    def test_normalize_result(self) -> None:
        cases = (
            (b"hello world", b"\n", b"hello world"),
            (b"hello\nworld\n", b"\n", b"hello\nworld\n"),
            (b"hello\r\nworld\r\n", b"\n", b"hello\r\nworld\r\n"),
            (b"hello world", b"\r\n", b"hello world"),
            (b"hello\nworld\n", b"\r\n", b"hello\r\nworld\r\n"),
        )
        for idx, (content, newline, expected) in enumerate(cases):
            with self.subTest(idx):
                self.assertEqual(expected, ufmt.util.normalize_result(content, newline))

    def test_read_file(self) -> None:
        with TemporaryDirectory() as td:
            tdp = Path(td).resolve()
            foo = tdp / "foo.py"

            with self.subTest("unix newlines"):
                foo.write_bytes(FAKE_CONTENT)

                result = ufmt.util.read_file(foo)
                expected = (FAKE_CONTENT, "utf-8", b"\n")
                self.assertTupleEqual(expected, result)

            with self.subTest("windows newlines"):
                content = FAKE_CONTENT.replace(b"\n", b"\r\n")
                foo.write_bytes(content)

                result = ufmt.util.read_file(foo)
                expected = (FAKE_CONTENT, "utf-8", b"\r\n")
                self.assertTupleEqual(expected, result)

            with self.subTest("empty file"):
                foo.write_bytes(b"")

                result = ufmt.util.read_file(foo)
                expected = (b"", "utf-8", b"\n")
                self.assertTupleEqual(expected, result)



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from typing import Any
from unittest import TestCase
from unittest.mock import ANY, Mock, patch

from trailrunner.tests.core import cd

from ufmt.config import load_config
from ufmt.types import UfmtConfig


class ConfigTest(TestCase):
    maxDiff = None

    def setUp(self) -> None:
        self._td = TemporaryDirectory()
        self.addCleanup(self._td.cleanup)

        self.td = Path(self._td.name).resolve()
        self.pyproject = self.td / "pyproject.toml"

    def subTest(self, *args: Any, **kwargs: Any) -> Any:
        load_config.cache_clear()
        return super().subTest(*args, **kwargs)

    def test_ufmt_config(self) -> None:
        fake_config = dedent("""
            [tool.ufmt]
            excludes = [
                "a",
                "b",
            ]
            """)

        foo = self.td / "foo"
        foo.mkdir()
        bar = foo / "bar.py"
        bar.write_text("print('hello world')\n")

        with self.subTest("absolute path, no pyproject.toml"):
            config = load_config(bar)
            self.assertEqual(UfmtConfig(), config)

        with self.subTest("local path, no pyproject.toml"):
            with cd(foo):
                config = load_config()
                self.assertEqual(UfmtConfig(), config)

        self.pyproject.write_text("")

        with self.subTest("absolute path, empty pyproject.toml"):
            config = load_config(bar)
            self.assertEqual(
                UfmtConfig(
                    project_root=self.td, pyproject_path=self.pyproject, excludes=[]
                ),
                config,
            )

        with self.subTest("local path, empty pyproject.toml"):
            with cd(self.td):
                config = load_config()
                self.assertEqual(
                    UfmtConfig(
                        project_root=self.td,
                        pyproject_path=self.pyproject,
                        excludes=[],
                    ),
                    config,
                )

        self.pyproject.write_text(fake_config)

        with self.subTest("absolute path, with pyproject.toml"):
            config = load_config(bar)
            self.assertEqual(
                UfmtConfig(
                    project_root=self.td,
                    pyproject_path=self.pyproject,
                    excludes=["a", "b"],
                ),
                config,
            )

        with self.subTest("local path, with pyproject.toml"):
            with cd(self.td):
                config = load_config()
                self.assertEqual(
                    UfmtConfig(
                        project_root=self.td,
                        pyproject_path=self.pyproject,
                        excludes=["a", "b"],
                    ),
                    config,
                )

        with self.subTest("absolute path, manually specify project root"):
            config = load_config(root=self.td)
            self.assertEqual(
                UfmtConfig(
                    project_root=self.td,
                    pyproject_path=self.pyproject,
                    excludes=["a", "b"],
                ),
                config,
            )

    @patch("ufmt.config.LOG")
    def test_invalid_config(self, log_mock: Mock) -> None:
        with self.subTest("string"):
            self.pyproject.write_text(dedent("""
                    [tool]
                    ufmt = "hello"
                    """))
            expected = UfmtConfig(project_root=self.td, pyproject_path=self.pyproject)
            result = load_config(self.td / "fake.py")
            self.assertEqual(expected, result)

            log_mock.warning.assert_called_once()
            log_mock.reset_mock()

        with self.subTest("array"):
            self.pyproject.write_text(dedent("""
                    [[tool.ufmt]]
                    excludes = ["fixtures/"]
                    """))
            expected = UfmtConfig(project_root=self.td, pyproject_path=self.pyproject)
            result = load_config(self.td / "fake.py")
            self.assertEqual(expected, result)

            log_mock.warning.assert_called_once()
            log_mock.reset_mock()

        with self.subTest("extra"):
            self.pyproject.write_text(dedent("""
                    [tool.ufmt]
                    unknown_element = true
                    hello_world = "my name is"
                    """))
            expected = UfmtConfig(project_root=self.td, pyproject_path=self.pyproject)
            result = load_config(self.td / "fake.py")
            self.assertEqual(expected, result)

            log_mock.warning.assert_called_with(
                ANY, self.pyproject, ["hello_world", "unknown_element"]
            )
            log_mock.reset_mock()

        with self.subTest("unsupported formatter"):
            self.pyproject.write_text(dedent("""
                    [tool.ufmt]
                    formatter = "garbage"
                    """))
            with self.assertRaisesRegex(
                ValueError, "'garbage' is not a valid Formatter"
            ):
                load_config(self.td / "fake.py")

        with self.subTest("unsupported sorter"):
            self.pyproject.write_text(dedent("""
                    [tool.ufmt]
                    sorter = "garbage"
                    """))
            with self.assertRaisesRegex(ValueError, "'garbage' is not a valid Sorter"):
                load_config(self.td / "fake.py")

    @patch("ufmt.config.LOG")
    def test_config_excludes(self, log_mock: Mock) -> None:
        with self.subTest("missing"):
            self.pyproject.write_text(dedent("""
                    [tool.ufmt]
                    """))
            expected = UfmtConfig(project_root=self.td, pyproject_path=self.pyproject)
            result = load_config(self.td / "fake.py")
            self.assertEqual(expected, result)
            log_mock.assert_not_called()

        with self.subTest("empty"):
            self.pyproject.write_text(dedent("""
                    [tool.ufmt]
                    excludes = []
                    """))
            expected = UfmtConfig(
                project_root=self.td, pyproject_path=self.pyproject, excludes=[]
            )
            result = load_config(self.td / "fake.py")
            self.assertEqual(expected, result)
            log_mock.assert_not_called()

        with self.subTest("list"):
            self.pyproject.write_text(dedent("""
                    [tool.ufmt]
                    excludes = ["fixtures/"]
                    """))
            expected = UfmtConfig(
                project_root=self.td,
                pyproject_path=self.pyproject,
                excludes=["fixtures/"],
            )
            result = load_config(self.td / "fake.py")
            self.assertEqual(expected, result)
            log_mock.assert_not_called()

        with self.subTest("string"):
            self.pyproject.write_text(dedent("""
                    [tool.ufmt]
                    excludes = "fixtures/"
                    """))
            with self.assertRaisesRegex(
                ValueError, "excludes must be a list of strings"
            ):
                load_config(self.td / "fake.py")



# Copyright 2022 Amethyst Reese, Tim Hatch
# Licensed under the MIT license

from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Callable, List, Optional, Union

from black import Mode as BlackConfig
from typing_extensions import Protocol
from usort import Config as UsortConfig

__all__ = [
    "BlackConfig",
    "BlackConfigFactory",
    "Encoding",
    "FileContent",
    "Newline",
    "Options",
    "Processor",
    "Result",
    "SkipFormatting",
    "STDIN",
    "UsortConfig",
    "UsortConfigFactory",
]

STDIN = Path("-")

Encoding = str
FileContent = bytes
Newline = bytes

BlackConfigFactory = Callable[[Path], BlackConfig]
UsortConfigFactory = Callable[[Path], UsortConfig]


class Formatter(Enum):
    """
    Select preferred formatter implementation.
    """

    black = "black"
    """Use black (default)."""

    ruff_api = "ruff-api"
    """
    **Experimental:**
    Use Ruff via unofficial `ruff-api <https://pypi.org/project/ruff-api>`_ extension.

    .. note::
        This implementation still depends on and uses the ``[tool.black]`` configuration
        table from ``pyproject.toml`` rather than Ruff's own configuration options.
        This may change in future updates.
    """


class Sorter(Enum):
    """
    Preferred import sorter implementation.
    """

    skip = "skip"
    """Skip sorting imports"""

    usort = "usort"
    """Use µsort (default)."""

    ruff_api = "ruff-api"
    """
    **Experimental:**
    Use Ruff's isort lint rules via unofficial
    `ruff-api <https://pypi.org/project/ruff-api>`_ extension.

    .. note::
        This implementation still depends on and uses the ``[tool.usort]`` configuration
        table from ``pyproject.toml`` rather than Ruff's own configuration options.
        This may change in future updates.
    """


@dataclass
class UfmtConfig:
    project_root: Optional[Path] = None
    pyproject_path: Optional[Path] = None
    excludes: List[str] = field(default_factory=list)
    formatter: Formatter = Formatter.black
    sorter: Sorter = Sorter.usort


UfmtConfigFactory = Callable[[Optional[Path], Optional[Path]], UfmtConfig]


@dataclass
class Options:
    debug: bool = False
    quiet: bool = False
    concurrency: Optional[int] = None
    root: Optional[Path] = None


class Processor(Protocol):
    def __call__(
        self,
        path: Path,
        content: FileContent,
        *,
        encoding: Encoding = "utf-8",
    ) -> FileContent:  # pragma: nocover
        """
        Process bytes before or after formatting and return the updated file content.
        """
        ...


class SkipFormatting(Exception):
    """Raise this exception in a pre/post processor to skip formatting a file."""


@dataclass
class Result:
    """
    Basic metadata results from formatting files.
    """

    path: Path
    changed: bool = False
    written: bool = False
    skipped: Union[bool, str] = False
    diff: Optional[str] = None
    error: Optional[Exception] = None
    before: bytes = b""  # only set if return_content=True
    after: bytes = b""  # only set if return_content=True



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

from .cli import CliTest
from .config import ConfigTest
from .core import CoreTest
from .lsp import LspTest
from .util import UtilTest



"""
Safe, atomic formatting with black and µsort
"""

__author__ = "Amethyst Reese"
from .__version__ import __version__
from .core import ufmt_bytes, ufmt_file, ufmt_paths, ufmt_stdin, ufmt_string
from .types import (
    BlackConfig,
    BlackConfigFactory,
    Encoding,
    FileContent,
    Formatter,
    Newline,
    Processor,
    Result,
    SkipFormatting,
    Sorter,
    UfmtConfig,
    UfmtConfigFactory,
    UsortConfig,
    UsortConfigFactory,
)

__all__ = [
    "__author__",
    "__version__",
    "BlackConfig",
    "BlackConfigFactory",
    "Encoding",
    "FileContent",
    "Formatter",
    "Newline",
    "Processor",
    "Result",
    "SkipFormatting",
    "Sorter",
    "ufmt_bytes",
    "ufmt_file",
    "ufmt_paths",
    "ufmt_stdin",
    "ufmt_string",
    "UfmtConfig",
    "UfmtConfigFactory",
    "UsortConfig",
    "UsortConfigFactory",
]



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

import os
import platform
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import skipIf, TestCase
from unittest.mock import call, Mock, patch

import trailrunner
from click.testing import CliRunner
from libcst import ParserSyntaxError

from ufmt.cli import echo_results, main
from ufmt.types import Result
from .core import CORRECTLY_FORMATTED_CODE, POORLY_FORMATTED_CODE


@patch.object(trailrunner.core.Trailrunner, "DEFAULT_EXECUTOR", ThreadPoolExecutor)
class CliTest(TestCase):
    def setUp(self) -> None:
        self.runner = CliRunner()
        self.cwd = os.getcwd()
        self.td = TemporaryDirectory()
        self.tdp = Path(self.td.name).resolve()
        os.chdir(self.tdp)

    def tearDown(self) -> None:
        os.chdir(self.cwd)
        self.td.cleanup()

    @patch("ufmt.cli.echo_color_precomputed_diff")
    @patch("ufmt.cli.click.secho")
    def test_echo(self, echo_mock: Mock, mol_mock: Mock) -> None:
        f1 = Path("foo/bar.py")
        f2 = Path("fuzz/buzz.py")
        f3 = Path("make/rake.py")
        results = [
            Result(f1, changed=False),
            Result(f2, changed=True, written=False, diff="fakediff1"),
            Result(f3, changed=True, written=True, diff="fakediff2"),
        ]

        with self.subTest("no diff"):
            echo_results(results)
            echo_mock.assert_has_calls(
                [
                    call(f"Would format {f2}", err=True),
                    call(f"Formatted {f3}", err=True),
                ]
            )
            mol_mock.assert_not_called()
            echo_mock.reset_mock()
            mol_mock.reset_mock()

        with self.subTest("with diff"):
            echo_results(results, diff=True)
            echo_mock.assert_has_calls(
                [
                    call(f"Would format {f2}", err=True),
                    call(f"Formatted {f3}", err=True),
                ]
            )
            mol_mock.assert_has_calls(
                [
                    call("fakediff1"),
                    call("fakediff2"),
                ]
            )
            echo_mock.reset_mock()
            mol_mock.reset_mock()

        with self.subTest("upstream exception"):
            results = [
                Result(f1, changed=False),
                Result(f2, error=AssertionError()),
                Result(f3, error=Exception("something weird happened")),
            ]
            echo_results(results)

            echo_mock.assert_has_calls(
                [
                    call(
                        f"Error formatting {f2}: AssertionError()",
                        fg="yellow",
                        err=True,
                    ),
                    call(
                        f"Error formatting {f3}: something weird happened",
                        fg="yellow",
                        err=True,
                    ),
                ]
            )
            mol_mock.assert_not_called()
            echo_mock.reset_mock()
            mol_mock.reset_mock()

    @patch("ufmt.cli.ufmt_paths")
    def test_check(self, ufmt_mock: Mock) -> None:
        with self.subTest("no paths given"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = []
            result = self.runner.invoke(main, ["check"])
            ufmt_mock.assert_called_with(
                [Path(".")], dry_run=True, concurrency=None, root=None
            )
            self.assertRegex(result.stderr, r"No files found")
            self.assertEqual(0, result.exit_code)

        with self.subTest("already formatted"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
                Result(Path("foo/frob.py"), changed=False),
            ]
            result = self.runner.invoke(main, ["check", "bar.py", "foo/frob.py"])
            ufmt_mock.assert_called_with(
                [Path("bar.py"), Path("foo/frob.py")],
                dry_run=True,
                concurrency=None,
                root=None,
            )
            self.assertEqual(0, result.exit_code)

        with self.subTest("needs formatting"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
                Result(Path("foo/frob.py"), changed=True),
            ]
            result = self.runner.invoke(main, ["check", "bar.py", "foo/frob.py"])
            ufmt_mock.assert_called_with(
                [Path("bar.py"), Path("foo/frob.py")],
                dry_run=True,
                concurrency=None,
                root=None,
            )
            self.assertEqual(1, result.exit_code)

        with self.subTest("syntax error"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
                Result(
                    Path("foo/frob.py"),
                    error=ParserSyntaxError(
                        "bad",
                        lines=("", "", "", "foo bar fizzbuzz hello world"),
                        raw_line=4,
                        raw_column=15,
                    ),
                ),
            ]
            result = self.runner.invoke(main, ["check", "bar.py", "foo/frob.py"])
            ufmt_mock.assert_called_with(
                [Path("bar.py"), Path("foo/frob.py")],
                dry_run=True,
                concurrency=None,
                root=None,
            )
            self.assertRegex(
                result.stderr, r"Error formatting .*frob\.py: Syntax Error @ 4:16"
            )
            self.assertEqual(1, result.exit_code)

        with self.subTest("skipped file"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("foo.py"), skipped="special"),
            ]
            result = self.runner.invoke(main, ["check", "foo.py"])
            ufmt_mock.assert_called_with(
                [Path("foo.py")], dry_run=True, concurrency=None, root=None
            )
            self.assertRegex(result.stderr, r"Skipped .*foo\.py: special")
            self.assertEqual(0, result.exit_code)

    @patch("ufmt.cli.ufmt_paths")
    def test_diff(self, ufmt_mock: Mock) -> None:
        with self.subTest("no paths given"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = []
            result = self.runner.invoke(main, ["diff"])
            ufmt_mock.assert_called_with(
                [Path(".")], dry_run=True, diff=True, concurrency=None, root=None
            )
            self.assertRegex(result.stderr, r"No files found")
            self.assertEqual(0, result.exit_code)

        with self.subTest("already formatted"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
                Result(Path("foo/frob.py"), changed=False),
            ]
            result = self.runner.invoke(main, ["diff", "bar.py", "foo/frob.py"])
            ufmt_mock.assert_called_with(
                [Path("bar.py"), Path("foo/frob.py")],
                dry_run=True,
                diff=True,
                concurrency=None,
                root=None,
            )
            self.assertEqual(0, result.exit_code)

        with self.subTest("needs formatting"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
                Result(Path("foo/frob.py"), changed=True),
            ]
            result = self.runner.invoke(main, ["diff", "bar.py", "foo/frob.py"])
            ufmt_mock.assert_called_with(
                [Path("bar.py"), Path("foo/frob.py")],
                dry_run=True,
                diff=True,
                concurrency=None,
                root=None,
            )
            self.assertEqual(1, result.exit_code)

        with self.subTest("syntax error"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
                Result(
                    Path("foo/frob.py"),
                    error=ParserSyntaxError(
                        "bad",
                        lines=("", "", "", "foo bar fizzbuzz hello world"),
                        raw_line=4,
                        raw_column=15,
                    ),
                ),
            ]
            result = self.runner.invoke(main, ["diff", "bar.py", "foo/frob.py"])
            ufmt_mock.assert_called_with(
                [Path("bar.py"), Path("foo/frob.py")],
                dry_run=True,
                diff=True,
                concurrency=None,
                root=None,
            )
            self.assertRegex(
                result.stderr, r"Error formatting .*frob\.py: Syntax Error @ 4:16"
            )
            self.assertEqual(1, result.exit_code)

        with self.subTest("skipped file"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("foo.py"), skipped="special"),
            ]
            result = self.runner.invoke(main, ["diff", "foo.py"])
            ufmt_mock.assert_called_with(
                [Path("foo.py")], dry_run=True, diff=True, concurrency=None, root=None
            )
            self.assertRegex(result.stderr, r"Skipped .*foo\.py: special")
            self.assertEqual(0, result.exit_code)

        with self.subTest("skipped file quiet"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("foo.py"), skipped="special"),
            ]
            result = self.runner.invoke(main, ["--quiet", "diff", "foo.py"])
            ufmt_mock.assert_called_with(
                [Path("foo.py")], dry_run=True, diff=True, concurrency=None, root=None
            )
            self.assertEqual("", result.stderr)
            self.assertEqual(0, result.exit_code)

        with self.subTest("bad root dir"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
            ]
            result = self.runner.invoke(
                main, ["--root", "DOES_NOT_EXIST", "diff", "bar.py"]
            )
            self.assertIn(
                r"Directory 'DOES_NOT_EXIST' does not exist",
                result.stderr,
            )
            self.assertEqual(2, result.exit_code)

    @patch("ufmt.cli.ufmt_paths")
    def test_format(self, ufmt_mock: Mock) -> None:
        with self.subTest("no paths given"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = []
            result = self.runner.invoke(main, ["format"])
            ufmt_mock.assert_called_with([Path(".")], concurrency=None, root=None)
            self.assertRegex(result.stderr, r"No files found")
            self.assertEqual(0, result.exit_code)

        with self.subTest("already formatted"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
                Result(Path("foo/frob.py"), changed=False),
            ]
            result = self.runner.invoke(main, ["format", "bar.py", "foo/frob.py"])
            ufmt_mock.assert_called_with(
                [Path("bar.py"), Path("foo/frob.py")], concurrency=None, root=None
            )
            self.assertEqual(0, result.exit_code)

        with self.subTest("needs formatting"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
                Result(Path("foo/frob.py"), changed=True),
            ]
            result = self.runner.invoke(main, ["format", "bar.py", "foo/frob.py"])
            ufmt_mock.assert_called_with(
                [Path("bar.py"), Path("foo/frob.py")], concurrency=None, root=None
            )
            self.assertEqual(0, result.exit_code)

        with self.subTest("syntax error"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("bar.py"), changed=False),
                Result(
                    Path("foo/frob.py"),
                    error=ParserSyntaxError(
                        "bad",
                        lines=("", "", "", "foo bar fizzbuzz hello world"),
                        raw_line=4,
                        raw_column=15,
                    ),
                ),
            ]
            result = self.runner.invoke(main, ["format", "bar.py", "foo/frob.py"])
            ufmt_mock.assert_called_with(
                [Path("bar.py"), Path("foo/frob.py")], concurrency=None, root=None
            )
            self.assertRegex(
                result.stderr, r"Error formatting .*frob\.py: Syntax Error @ 4:16"
            )
            self.assertRegex(result.stderr, " 1 error, 1 file already formatted")
            self.assertEqual(1, result.exit_code)

        with self.subTest("skipped file"):
            ufmt_mock.reset_mock()
            ufmt_mock.return_value = [
                Result(Path("foo.py"), skipped="special"),
            ]
            result = self.runner.invoke(main, ["format", "foo.py"])
            ufmt_mock.assert_called_with([Path("foo.py")], concurrency=None, root=None)
            self.assertRegex(result.stderr, r"Skipped .*foo\.py: special")
            self.assertEqual(0, result.exit_code)

    @skipIf(platform.system() == "Windows", "stderr not supported on Windows")
    def test_stdin(self) -> None:
        with self.subTest("check clean"):
            result = self.runner.invoke(
                main,
                ["check", "-", "hello.py"],
                input=CORRECTLY_FORMATTED_CODE,
            )
            self.assertEqual("", result.stdout)
            self.assertRegex(result.stderr, r"✨ 1 file already formatted ✨")
            self.assertEqual(0, result.exit_code)

        with self.subTest("check dirty"):
            result = self.runner.invoke(
                main,
                ["check", "-"],
                input=POORLY_FORMATTED_CODE,
            )
            self.assertEqual("", result.stdout)
            self.assertIn("Would format stdin\n", result.stderr)
            self.assertEqual(1, result.exit_code)

        with self.subTest("diff clean"):
            result = self.runner.invoke(
                main,
                ["diff", "-", "hello.py"],
                input=CORRECTLY_FORMATTED_CODE,
            )
            self.assertEqual("", result.stdout)
            self.assertIn("✨ 1 file already formatted ✨", result.stderr)
            self.assertEqual(0, result.exit_code)

        with self.subTest("diff dirty"):
            result = self.runner.invoke(
                main,
                ["diff", "-", "hello.py"],
                input=POORLY_FORMATTED_CODE,
            )
            self.assertRegex(result.stdout, r"--- hello.py\n\+\+\+ hello.py")
            self.assertIn("Would format hello.py\n", result.stderr)
            self.assertEqual(1, result.exit_code)

        with self.subTest("format clean"):
            result = self.runner.invoke(
                main,
                ["format", "-", "hello.py"],
                input=CORRECTLY_FORMATTED_CODE,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE, result.stdout)
            self.assertIn("✨ 1 file already formatted ✨", result.stderr)
            self.assertEqual(0, result.exit_code)

        with self.subTest("format dirty"):
            result = self.runner.invoke(
                main,
                ["format", "-", "hello.py"],
                input=POORLY_FORMATTED_CODE,
            )
            self.assertEqual(CORRECTLY_FORMATTED_CODE, result.stdout)
            self.assertIn("Formatted hello.py\n", result.stderr)
            self.assertEqual(0, result.exit_code)

    def test_end_to_end(self) -> None:
        alpha = self.tdp / "alpha.py"
        beta = self.tdp / "beta.py"
        (self.tdp / "sub").mkdir()
        gamma = self.tdp / "sub" / "gamma.py"
        kappa = self.tdp / "sub" / "kappa.py"

        def reset() -> None:
            alpha.write_text(CORRECTLY_FORMATTED_CODE)
            beta.write_text(POORLY_FORMATTED_CODE)
            gamma.write_text(CORRECTLY_FORMATTED_CODE)
            kappa.write_text(POORLY_FORMATTED_CODE)

        with self.subTest("check"):
            reset()
            result = self.runner.invoke(main, ["check", self.tdp.as_posix()])
            self.assertEqual("", result.stdout)
            self.assertIn(f"Would format {beta}", result.stderr)
            self.assertIn(f"Would format {kappa}", result.stderr)
            self.assertIn(
                "2 files would be formatted, 2 files already formatted", result.stderr
            )
            self.assertEqual(POORLY_FORMATTED_CODE, beta.read_text())
            self.assertEqual(POORLY_FORMATTED_CODE, kappa.read_text())

        with self.subTest("diff"):
            reset()
            result = self.runner.invoke(main, ["diff", self.tdp.as_posix()])
            self.assertIn(f"--- {beta.as_posix()}", result.stdout)
            self.assertIn(f"+++ {beta.as_posix()}", result.stdout)
            self.assertIn(f"Would format {beta}", result.stderr)
            self.assertIn(f"--- {kappa.as_posix()}", result.stdout)
            self.assertIn(f"+++ {kappa.as_posix()}", result.stdout)
            self.assertIn(f"Would format {kappa}", result.stderr)
            self.assertIn(
                "2 files would be formatted, 2 files already formatted", result.stderr
            )
            self.assertEqual(POORLY_FORMATTED_CODE, beta.read_text())
            self.assertEqual(POORLY_FORMATTED_CODE, kappa.read_text())

        with self.subTest("format"):
            reset()
            result = self.runner.invoke(main, ["format", self.tdp.as_posix()])
            self.assertEqual("", result.stdout)
            self.assertIn(f"Formatted {beta}", result.stderr)
            self.assertIn(f"Formatted {kappa}", result.stderr)
            self.assertIn("2 files formatted, 2 files already formatted", result.stderr)
            self.assertEqual(CORRECTLY_FORMATTED_CODE, beta.read_text())
            self.assertEqual(CORRECTLY_FORMATTED_CODE, kappa.read_text())

        with self.subTest("format quiet"):
            reset()
            result = self.runner.invoke(
                main, ["--quiet", "format", self.tdp.as_posix()]
            )
            self.assertEqual("", result.stdout)
            self.assertEqual("", result.stderr)
            self.assertEqual(CORRECTLY_FORMATTED_CODE, beta.read_text())
            self.assertEqual(CORRECTLY_FORMATTED_CODE, kappa.read_text())

        with self.subTest("format subdir"):
            reset()
            result = self.runner.invoke(main, ["format", (self.tdp / "sub").as_posix()])
            self.assertEqual("", result.stdout)
            self.assertNotIn(f"Formatted {beta}", result.stderr)
            self.assertIn(f"Formatted {kappa}", result.stderr)
            self.assertIn("1 file formatted, 1 file already formatted", result.stderr)
            self.assertEqual(POORLY_FORMATTED_CODE, beta.read_text())
            self.assertEqual(CORRECTLY_FORMATTED_CODE, kappa.read_text())

    @patch("ufmt.lsp.ufmt_lsp")  # dynamic import, patch at definition
    def test_lsp(self, lsp_mock: Mock) -> None:
        with self.subTest("default"):
            lsp_mock.reset_mock()
            self.runner.invoke(main, ["lsp"])

            lsp_mock.assert_called_with(root=None)
            lsp_mock.return_value.start_io.assert_called_with()

        with self.subTest("tcp"):
            lsp_mock.reset_mock()
            self.runner.invoke(main, ["lsp", "--tcp", "--port", "4567"])

            lsp_mock.assert_called_with(root=None)
            lsp_mock.return_value.start_tcp.assert_called_with("localhost", 4567)

        with self.subTest("ws"):
            lsp_mock.reset_mock()
            self.runner.invoke(main, ["lsp", "--ws", "--port", "8765"])

            lsp_mock.assert_called_with(root=None)
            lsp_mock.return_value.start_ws.assert_called_with("localhost", 8765)



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

import unittest

from ufmt.cli import init_logging

if __name__ == "__main__":  # pragma: no cover
    init_logging()
    unittest.main(module="ufmt.tests", verbosity=2)



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license
import logging
from functools import lru_cache
from pathlib import Path
from typing import Optional, Sequence

import tomlkit
from trailrunner import project_root

from .types import Formatter, Sorter, UfmtConfig

LOG = logging.getLogger(__name__)


@lru_cache
def load_config(path: Optional[Path] = None, root: Optional[Path] = None) -> UfmtConfig:
    path = path or Path.cwd()
    if root is None:
        root = project_root(path)
    config_path = root / "pyproject.toml"
    if config_path.is_file():
        pyproject = tomlkit.loads(config_path.read_text())
        config = pyproject.get("tool", {}).get("ufmt", {})
        if not isinstance(config, dict):
            LOG.warning("%s: tool.ufmt is not a mapping, ignoring", config_path)
            config = {}

        config_excludes = config.pop("excludes", [])
        if isinstance(config_excludes, Sequence) and not isinstance(
            config_excludes, str
        ):
            excludes = [str(x) for x in config_excludes]
        else:
            raise ValueError(f"{config_path}: excludes must be a list of strings")

        formatter = Formatter(config.pop("formatter", UfmtConfig.formatter))
        sorter = Sorter(config.pop("sorter", Sorter.usort))

        if config:
            LOG.warning("%s: unknown values ignored: %r", config_path, sorted(config))

        return UfmtConfig(
            project_root=root,
            pyproject_path=config_path,
            excludes=excludes,
            formatter=formatter,
            sorter=sorter,
        )

    return UfmtConfig()
